#include <ygUtil.h>
#include <ygWorld.h>
#include <ygNodeDB.h>
#include "flyingCarpet.h"

using namespace std;

extern "C" ygNode* construct_flyingCarpet(const char* name,bool master) { return new flyingCarpet(name,master); }

struct _flyingCarpetPrivateData
	{
	float maxHeight;
	bool active;
	ygString userName;
	pfVec3 startPos;
	float startHeading;
	pfVec3 destPos;
	float destHeading;
	float duration;
	float startTime;
	};


flyingCarpet::flyingCarpet(const char* name,bool master) : ygTransform(name,master)
	{
	setClassName("flyingCarpet");
	p_ = new struct _flyingCarpetPrivateData;
	p_->maxHeight = 100;
	p_->active = false;
	}

flyingCarpet::~flyingCarpet(void)
	{
	delete p_;
	}


void flyingCarpet::reset(void)
	{
	p_->maxHeight = 100;
	p_->active = false;
	ygTransform::reset();
	}


void flyingCarpet::message(const ygMessage& msg)
	{
	if (msg == "move")
		{
		if (msg.args.size() != 6)
			{
			msg.error(name(),"(wrong number of arguments - should be "
				"username dest-x dest-y dest-z dest-heading time)");
			return;
			}
		pfVec3 destination;
		msg.getVec3Args(destination,1);
		float destHeading = msg.floatArg(4);
		float duration = msg.floatArg(5);
		move(msg.args[0],destination,destHeading,duration);
		}
	else if (msg == "maxheight")
		p_->maxHeight = msg.floatArg(0);
	else
		ygTransform::message(msg);
	}


void flyingCarpet::move(const ygString& userName,const pfVec3& destination,
			float destHeading,float duration)
	{
	ygNode * user = ygNodeDB::find(userName,false);
	if (!user)
		return;
	p_->active = true;
	p_->startTime = ygWorld::FrameTime;
	p_->userName = userName;
	pfMatrix userMat, parentMat, invParentMat;
	user->getTransform(userMat);
	getParentTransform(parentMat);
	invParentMat.invertAff(parentMat);
	userMat *= invParentMat;
	pfVec3 origin(0,0,0), forward(0,1,0), userDir;
	p_->startPos.xformPt(origin,userMat);
	userDir.xformVec(forward,userMat);
	p_->startHeading = pfArcTan2(-userDir[0],userDir[1]);
	p_->destPos = destination;
	p_->destHeading = destHeading;
	if (duration < 0.00001)
		duration = 0.00001;
	p_->duration = duration;
	setPosition(p_->startPos);
	setOrientation(0,0,p_->startHeading);
	ygString msgStr(userName);
	msgStr += ".attach(";
	msgStr += name();
	msgStr += ")";
	ygMessage msg;
	msg.parseString(msgStr);
	ygWorld::World->scheduleMessage(msg);
	}


void flyingCarpet::stop(void)
	{
	ygString msgStr(p_->userName);
	msgStr += ".release";
	ygMessage msg;
	msg.parseString(msgStr);
	ygWorld::World->scheduleMessage(msg);
	p_->active = false;
	}


void flyingCarpet::app(void)
	{
	if (p_->active)
		{
		float t = (ygWorld::FrameTime - p_->startTime) / p_->duration;
		if (t > 1.0)
			t = 1.0;
		float a = 3*t*t - 2*t*t*t;
		interpolate(a);
		if (t >= 1.0)
			stop();
		}
	ygTransform::app();
	}


void flyingCarpet::interpolate(float a)
	{
	pfVec3 pos;
	float heading, A, B, C;
	pos = (1.0f - a) * p_->startPos + a * p_->destPos;
	A = 2*p_->startPos[2] + 2*p_->destPos[2] - 4*p_->maxHeight;
	B = -3*p_->startPos[2] - p_->destPos[2] + 4*p_->maxHeight;
	C = p_->startPos[2];
	pos[2] = A*a*a + B*a + C;
	heading = (1.0f - a) * p_->startHeading + a * p_->destHeading;
	setPosition(pos);
	setOrientation(0,0,heading);
	}
