#include <stdio.h>
#include <bergenSample.h>
#include <bergenServer.h>
#include <vector>
#include "ygVolume.h"
#include "ygUtil.h"
#include "ygWorld.h"
#include "ygNetKeys.h"
#include "ygSound.h"

using namespace std;

extern "C" ygNode* construct_ygSound(const char* name,bool master) { return new ygSound(name,master); }

struct _ygSoundPrivateData
	{
	ygString file;
	bergenSample * sample;
	float duration;
	float falloffDistance;
	float amplitude;
	bool loop;
	float startTime, endTime;
	float curAmplitude;
	bool isPlaying;
	unsigned int lastHeartbeat;
	ygString netFile;
	bool netPlay;
	};


struct _ygSoundGlobalData
	{
	bergenServer * server;
	vector<ygSound*> sounds;
	unsigned int currentHeartbeat;
	};

struct _ygSoundGlobalData * ygSound::global_ = NULL;



ygSound::ygSound(const char* name,bool master) : ygSpace(name,master)
	{
	setClassName("ygSound",true);
	if (!global_)
		initServer();
	p_ = new struct _ygSoundPrivateData;
	p_->sample = NULL;
	p_->falloffDistance = 10;
	p_->amplitude = 1;
	p_->loop = false;
	p_->startTime = 0;
	p_->endTime = 0;
	p_->curAmplitude = 0;
	p_->isPlaying = false;
//	debugFlag("myDebug", &p_->myDebug);
	addSound(this);
	addNetKey("file", &p_->netFile, YG_NET_STRING);
	addNetKey("ampl", &p_->amplitude, YG_NET_FLOAT);
	addNetKey("dist", &p_->falloffDistance, YG_NET_FLOAT);
	addNetKey("loop", &p_->loop, YG_NET_BOOL);
	addNetKey("play", &p_->netPlay, YG_NET_BOOL);
	}

ygSound::~ygSound(void)
	{
	if (p_->sample)
		delete p_->sample;
	delete p_;
	}


void ygSound::reset(void)
	{
	if (p_->sample)
		{
		stop();
		delete p_->sample;
		p_->sample = NULL;
		}
	ygSpace::reset();
	}


void ygSound::message(const ygMessage& msg)
	{
	if (msg == "File")
		{
		if (msg.args.size() > 0)
			setFile(msg.args[0]);
		}
	else if (msg == "FalloffDistance")
		setFalloffDistance(msg.floatArg(0));
	else if (msg == "Amplitude")
		setAmplitude(msg.floatArg(0));
	else if (msg == "Loop")
		setLoop(msg.boolArg(0));
	else if (msg == "Play")
		{
		if (msg.args.size() > 0)
			setFile(msg.args[0]);
		if (msg.args.size() > 1)
			setAmplitude(msg.floatArg(1));
		play();
		}
	else if (msg == "Stop")
		stop();
	else
		ygSpace::message(msg);
	}


void ygSound::setFile(const ygString& val)
	{
	if (p_->sample)
		{
		stop();
		delete p_->sample;
		p_->sample = NULL;
		}
	p_->file = val;
	p_->netFile = p_->file;
	netKeyChanged("file");
	}


void ygSound::setFalloffDistance(float val)
	{
	p_->falloffDistance = val;
	netKeyChanged("dist");
	}


void ygSound::setAmplitude(float val)
	{
	p_->amplitude = val;
	netKeyChanged("ampl");
	}


void ygSound::setLoop(bool val)
	{
	p_->loop = val;
	if (p_->sample)
		p_->sample->setLoop(val);
	netKeyChanged("loop");
	}


void ygSound::setCurAmplitude(float val)
	{
	p_->curAmplitude = val;
	if (p_->sample)
		p_->sample->setAmplitude(val);
	}


bool ygSound::isPlaying(void)
	{
	return p_->isPlaying;
	}


void ygSound::play(void)
	{
	if (p_->isPlaying)
		return;
	if (!p_->sample)
		{
		p_->sample = new bergenSample(p_->file.c_str(),server());
		p_->duration = p_->sample->duration();
		if (p_->loop)
			p_->sample->setLoop(1);
		else
			p_->sample->setLoop(0);
		}
	p_->sample->setLoop(p_->loop);
	setCurAmplitude(computeAmplitude());
	p_->sample->play();
	heartbeat();
	p_->isPlaying = true;
	if (!p_->loop)
		{
		p_->startTime = ygWorld::FrameTime;
		p_->endTime = p_->startTime + p_->duration;
		}
	eventOccurred("StartPlaying");
	p_->netPlay = true;
	netKeyChanged("play");
	}


void ygSound::stop(bool sendStopCommand)
	{
	if (p_->isPlaying)
		{
		if ((sendStopCommand) && (p_->sample))
			{
			p_->sample->stop();
			setCurAmplitude(0.0);
			}
		p_->isPlaying = false;
		eventOccurred("StopPlaying");
		p_->netPlay = false;
		netKeyChanged("play");
		}
	}


void ygSound::heartbeat(void)
	{
	if (global_)
		p_->lastHeartbeat = global_->currentHeartbeat;
	}


void ygSound::app(void)
	{
	if (p_->sample)
		{
		heartbeat();
		if (p_->isPlaying)
			{
			if ((!p_->loop) && (p_->endTime < ygWorld::FrameTime))
				stop(false);
			else
				{
				float newAmpl = computeAmplitude();
				if (fabs(newAmpl - p_->curAmplitude) > 0.005)
					setCurAmplitude(newAmpl);
				}
			}
		}
	ygSpace::app();
	}


float ygSound::computeAmplitude(void)
	{
	float amp = p_->amplitude;
	float maxDist = p_->falloffDistance;
	ygNode * viewer = ygWorld::World->viewer();
	if ((maxDist >= 0) && (viewer))
		{
		viewer->app();
		pfVec3 userPos = viewer->origin(this);
		float distSq;
		if (volume())
			distSq = volume()->sqrDistance(userPos);
		else
			distSq = userPos.dot(userPos);
		if (distSq > maxDist*maxDist)
			amp = 0;
		else if (maxDist > 0)
			amp *= (1.0f - sqrtf(distSq) / maxDist);
		}
	return amp;
	}


void ygSound::acceptNetKey(const ygString& key)
	{
	if (key == "file")
		setFile(p_->netFile);
	else if (key == "play")
		{
		if (p_->netPlay)
			play();
		else
			stop();
		}
	else
		ygSpace::acceptNetKey(key);
	}


/****** Global functions for connection to server & heartbeat *******/

void ygSound::initServer(void)
	{
	if (!global_)
		{
		global_ = new struct _ygSoundGlobalData;
		global_->server = new bergenServer;
		if (getenv("YG_SOUND_DIR"))
			global_->server->setDirectory(getenv("YG_SOUND_DIR"));
		global_->currentHeartbeat = 0;
		}
	}


void ygSound::addSound(ygSound* s)
	{
	if (global_)
		global_->sounds.push_back(s);
	}


void ygSound::removeSound(ygSound* s)
	{
	if (global_)
		{
		vector<ygSound*>::iterator iter;
		for (iter = global_->sounds.begin(); iter != global_->sounds.end();
		    ++iter)
			if (*iter == s)
				{
				global_->sounds.erase(iter);
				break;
				}
		}
	}


void ygSound::updateServer(void)
	{
	vector<ygSound*>::iterator iter;
	for (iter = global_->sounds.begin(); iter != global_->sounds.end(); ++iter)
		if (((*iter)->p_->isPlaying) &&
		    ((*iter)->p_->lastHeartbeat < global_->currentHeartbeat))
			(*iter)->stop();
	global_->currentHeartbeat++;
	}


void ygSound::closeServer(void)
	{
	delete global_->server;
	delete global_;
	global_ = NULL;
	}


bergenServer * ygSound::server(void)
	{
	return global_->server;
	}
