#include <stdlib.h>
#include <iostream>
#include <list>
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfEarthSky.h>
#include <Performer/pr/pfFog.h>
#include <Performer/pf/pfScene.h>
#include <pfcave.h>
#include "ygUtil.h"
#include "ygNodeDB.h"
#include "ygNet.h"
#include "ygDebugFlags.h"
#include "ygSound.h"
#include "ygCAVEViewer.h"
#include "ygWorld.h"

using namespace std;


class ygDelayMsg
	{
	public:
	 ygMessage msg;
	 float time;
	 friend inline bool operator< (const ygDelayMsg& a,const ygDelayMsg& b)
			{ return a.time < b.time; };
	};


struct _ygWorldPrivateData
	{
	bool sentMessageDebug;
	list<ygDelayMsg*> messageQueue;
	pfScene* pfscene;
	pfEarthSky* pfsky;
	pfFog* pffog;
	vector<ygUser*> users;
	ygNode* viewer;
	};



float ygWorld::FrameTime = 0;
float ygWorld::FrameDeltaTime = 1.0;
int ygWorld::FrameNumber = 0;
ygWorld* ygWorld::World = NULL;


ygWorld::ygWorld(void)
	{
	ygString dummyName;
	if (World)
		cerr << "ERROR: More than one ygWorld has been created -"
			" there should be only one!\n";
	World = this;
	p_ = new struct _ygWorldPrivateData;
	p_->pfscene = new pfScene;
	p_->pfsky = new pfEarthSky;
	p_->pffog = new pfFog;
	pfCAVEMasterChan()->setESky(p_->pfsky);
	p_->sentMessageDebug = ygDebugFlags::checkDebugEnv("world.sentmessages");
	ygNode::createDummyName(dummyName);
	p_->viewer = new ygCAVEViewer(dummyName.c_str(),true);
	ygNode::createDummyName(dummyName);
	localRoot_ = ygNodeDB::create("group",dummyName);
	ygSound::initServer();
	}


ygWorld::~ygWorld(void)
	{
	while (!p_->messageQueue.empty())
		{
		ygDelayMsg *dm = p_->messageQueue.front();
		p_->messageQueue.pop_front();
		delete dm;
		}
	delete p_;
	ygSound::closeServer();
	}


void ygWorld::createRoot(void)
	{
	root_ = ygNodeDB::create("group","root");
	root_->addChild(localRoot_);
	p_->pfscene->addChild(root_->pfnode());
	}


void ygWorld::setRoot(ygNode *r)
	{
	root_ = r;
	if (root_)
		{
		root_->addChild(localRoot_);
		p_->pfscene->addChild(root_->pfnode());
		}
	}


pfScene* ygWorld::pfscene(void) const
	{
	return p_->pfscene;
	}


ygNode* ygWorld::viewer(void) const
	{
	return p_->viewer;
	}


void ygWorld::reset(void)
	{
	FrameTime = ygGetTime();
	while (!p_->messageQueue.empty())
		{
		ygDelayMsg *dm = p_->messageQueue.front();
		p_->messageQueue.pop_front();
		delete dm;
		}
	if (localRoot_)
		localRoot_->resetTree();
	}


void ygWorld::frame(void)
	{
	float t = ygGetTime();
	FrameDeltaTime = t - FrameTime;
	FrameTime = t;
	if (FrameDeltaTime <= 0)
		{
		cerr << "ERROR (ygWorld::app): Bogus delta time ("
			<< FrameDeltaTime << "); kludging to .01\n";
		FrameDeltaTime = 0.01;
		}
	FrameNumber++;
	while ((!p_->messageQueue.empty()) &&
		(p_->messageQueue.front()->time <= FrameTime))
		{
		ygDelayMsg *dm = p_->messageQueue.front();
		sendMessage(dm->msg);
		p_->messageQueue.pop_front();
		delete dm;
		}
	if (root_)
		root_->appTraverse();
	if (p_->viewer)
		p_->viewer->app();
	ygSound::updateServer();
	ygNet::process();
	}


void ygWorld::scheduleMessage(ygMessage& msg)
	{
	if (msg.delay <= 0)
		sendMessage(msg);
	else
		{
		ygDelayMsg * dm = new ygDelayMsg;
		dm->time = FrameTime + msg.delay;
		dm->msg = msg;
		list<ygDelayMsg*>::iterator i = p_->messageQueue.begin();
		for ( ; i != p_->messageQueue.end(); ++i)
			if (dm->time < (*i)->time)
				break;
		p_->messageQueue.insert(i,dm);
		}
	}


void ygWorld::sendMessage(ygMessage& msg)
	{
	if (p_->sentMessageDebug)
		cout << "ygWorld: sending message " << msg << endl;
	if (!msg.node)
		msg.node = ygNodeDB::find(msg.nodeName,false);
	if (msg.node)
		{
		if (msg.node->isNetMaster())
			msg.node->message(msg);
		else
			msg.node->sendNetMessage(msg);
		}
	else
		cerr << "ERROR: could not find recipient node for message '" << msg << "'\n";
	}


void ygWorld::addUser(ygUser* user)
	{
	p_->users.push_back(user);
	}


void ygWorld::removeUser(ygUser* user)
	{
	vector<ygUser*>::iterator i = p_->users.begin();
	while ((i != p_->users.end()) && (*i != user))
		++i;
	if (i != p_->users.end())
		p_->users.erase(i);
	}


int ygWorld::numUsers(void) const
	{
	return p_->users.size();
	}


ygUser* ygWorld::user(int index) const
	{
	if ((index >= 0) && (index < p_->users.size()))
		return p_->users[index];
	else
		return NULL;
	}


void ygWorld::setClip(float near,float far)
	{
	CAVENear = near;
	CAVEFar = far;
	}


void ygWorld::setSkyColor(const pfVec3& sky)
	{
	p_->pfsky->setColor(PFES_CLEAR, sky[0], sky[1], sky[2], 0.0);
	}


void ygWorld::fogOff(void)
	{
	p_->pfsky->setFog(PFES_GENERAL, NULL);
	}


void ygWorld::setFog(int fogtype,const pfVec3& color,float onset,float opaque)
	{
	p_->pffog->setFogType(fogtype);
	p_->pffog->setColor(color[0],color[1],color[2]);
	p_->pffog->setRange(onset,opaque);
	p_->pfsky->setFog(PFES_GENERAL, p_->pffog);
	}


void ygWorld::unlinkLocalScene(void)
	{
	unlinkTree(localRoot());
	frame();
	}


void ygWorld::unlinkTree(ygNode *node)
	{
	if (!node)
		return;
	while (node->numChildren() > 0)
		unlinkTree(node->child(0));
	if (node->parent())
		node->parent()->removeChild(node);
	}
